@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE NondecreasingIndentation #-}
 {-# LANGUAGE OverloadedStrings #-}
@@ -18,13 +18,14 @@
 
 module GHC.Driver.Backpack (doBackpack) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
+import GHC.Driver.Backend
 -- In a separate module because it hooks into the parser.
 import GHC.Driver.Backpack.Syntax
-import GHC.Driver.Config
+import GHC.Driver.Config.Finder (initFinderOpts)
+import GHC.Driver.Config.Parser (initParserOpts)
+import GHC.Driver.Config.Diagnostic
 import GHC.Driver.Monad
 import GHC.Driver.Session
 import GHC.Driver.Ppr
@@ -32,12 +33,14 @@ import GHC.Driver.Main
 import GHC.Driver.Make
 import GHC.Driver.Env
 import GHC.Driver.Errors
+import GHC.Driver.Errors.Types
 
 import GHC.Parser
 import GHC.Parser.Header
 import GHC.Parser.Lexer
 import GHC.Parser.Annotation
-import GHC.Parser.Errors.Ppr
+
+import GHC.Rename.Names
 
 import GHC hiding (Failed, Succeeded)
 import GHC.Tc.Utils.Monad
@@ -46,13 +49,13 @@ import GHC.Builtin.Names
 
 import GHC.Types.SrcLoc
 import GHC.Types.SourceError
-import GHC.Types.SourceText
 import GHC.Types.SourceFile
 import GHC.Types.Unique.FM
 import GHC.Types.Unique.DFM
 import GHC.Types.Unique.DSet
 
 import GHC.Utils.Outputable
+import GHC.Utils.Fingerprint
 import GHC.Utils.Misc
 import GHC.Utils.Panic
 import GHC.Utils.Error
@@ -61,7 +64,6 @@ import GHC.Utils.Logger
 import GHC.Unit
 import GHC.Unit.Env
 import GHC.Unit.External
-import GHC.Unit.State
 import GHC.Unit.Finder
 import GHC.Unit.Module.Graph
 import GHC.Unit.Module.ModSummary
@@ -72,6 +74,7 @@ import GHC.Linker.Types
 import qualified GHC.LanguageExtensions as LangExt
 
 import GHC.Data.Maybe
+import GHC.Data.OsPath (unsafeEncodeUtf, os)
 import GHC.Data.StringBuffer
 import GHC.Data.FastString
 import qualified GHC.Data.EnumSet as EnumSet
@@ -88,27 +91,32 @@ import Data.IORef
 import Data.Map (Map)
 import qualified Data.Map as Map
 import qualified Data.Set as Set
+import GHC.Types.Error (mkUnknownDiagnostic)
 
 -- | Entry point to compile a Backpack file.
 doBackpack :: [FilePath] -> Ghc ()
 doBackpack [src_filename] = do
-    logger <- getLogger
-
     -- Apply options from file to dflags
     dflags0 <- getDynFlags
     let dflags1 = dflags0
-    src_opts <- liftIO $ getOptionsFromFile dflags1 src_filename
+    let parser_opts1 = initParserOpts dflags1
+    (p_warns, src_opts) <- liftIO $ getOptionsFromFile parser_opts1 src_filename
     (dflags, unhandled_flags, warns) <- liftIO $ parseDynamicFilePragma dflags1 src_opts
-    modifySession (\hsc_env -> hsc_env {hsc_dflags = dflags})
+    modifySession (hscSetFlags dflags)
+    logger <- getLogger -- Get the logger after having set the session flags,
+                        -- so that logger options are correctly set.
+                        -- Not doing so caused #20396.
     -- Cribbed from: preprocessFile / GHC.Driver.Pipeline
     liftIO $ checkProcessArgsResult unhandled_flags
-    liftIO $ handleFlagWarnings logger dflags warns
+    let print_config = initPrintConfig dflags
+    liftIO $ printOrThrowDiagnostics logger print_config (initDiagOpts dflags) (GhcPsMessage <$> p_warns)
+    liftIO $ printOrThrowDiagnostics logger print_config (initDiagOpts dflags) (GhcDriverMessage <$> warns)
     -- TODO: Preprocessing not implemented
 
     buf <- liftIO $ hGetStringBuffer src_filename
     let loc = mkRealSrcLoc (mkFastString src_filename) 1 1 -- TODO: not great
     case unP parseBackpack (initParserState (initParserOpts dflags) buf loc) of
-        PFailed pst -> throwErrors (fmap pprError (getErrorMessages pst))
+        PFailed pst -> throwErrors (GhcPsMessage <$> getPsErrorMessages pst)
         POk _ pkgname_bkp -> do
             -- OK, so we have an LHsUnit PackageName, but we want an
             -- LHsUnit HsComponentId.  So let's rename it.
@@ -121,21 +129,21 @@ doBackpack [src_filename] = do
                     innerBkpM $ do
                         let (cid, insts) = computeUnitId lunit
                         if null insts
-                            then if cid == Indefinite (UnitId (fsLit "main"))
+                            then if cid == UnitId (fsLit "main")
                                     then compileExe lunit
                                     else compileUnit cid []
                             else typecheckUnit cid insts
 doBackpack _ =
     throwGhcException (CmdLineError "--backpack can only process a single file")
 
-computeUnitId :: LHsUnit HsComponentId -> (IndefUnitId, [(ModuleName, Module)])
+computeUnitId :: LHsUnit HsComponentId -> (UnitId, [(ModuleName, Module)])
 computeUnitId (L _ unit) = (cid, [ (r, mkHoleModule r) | r <- reqs ])
   where
     cid = hsComponentId (unLoc (hsunitName unit))
     reqs = uniqDSetToList (unionManyUniqDSets (map (get_reqs . unLoc) (hsunitBody unit)))
-    get_reqs (DeclD HsigFile (L _ modname) _) = unitUniqDSet modname
     get_reqs (DeclD HsSrcFile _ _) = emptyUniqDSet
     get_reqs (DeclD HsBootFile _ _) = emptyUniqDSet
+    get_reqs (DeclD HsigFile (L _ modname) _) = unitUniqDSet modname
     get_reqs (IncludeD (IncludeDecl (L _ hsuid) _ _)) =
         unitFreeModuleHoles (convertHsComponentId hsuid)
 
@@ -154,7 +162,7 @@ data SessionType
 
 -- | Create a temporary Session to do some sort of type checking or
 -- compilation.
-withBkpSession :: IndefUnitId
+withBkpSession :: UnitId
                -> [(ModuleName, Module)]
                -> [(Unit, ModRenaming)]
                -> SessionType   -- what kind of session are we doing
@@ -162,7 +170,7 @@ withBkpSession :: IndefUnitId
                -> BkpM a
 withBkpSession cid insts deps session_type do_this = do
     dflags <- getDynFlags
-    let cid_fs = unitFS (indefUnit cid)
+    let cid_fs = unitFS cid
         is_primary = False
         uid_str = unpackFS (mkInstantiatedUnitHash cid insts)
         cid_str = unpackFS cid_fs
@@ -180,18 +188,20 @@ withBkpSession cid insts deps session_type do_this = do
                  , not (null insts) = sub_comp (key_base p) </> uid_str
                  | otherwise = sub_comp (key_base p)
 
-        mk_temp_env hsc_env = hsc_env
-            { hsc_dflags   = mk_temp_dflags (hsc_units hsc_env) (hsc_dflags hsc_env)
-            }
+        mk_temp_env hsc_env =
+          hscUpdateFlags (\dflags -> mk_temp_dflags (hsc_units hsc_env) dflags) hsc_env
         mk_temp_dflags unit_state dflags = dflags
             { backend = case session_type of
-                            TcSession -> NoBackend
+                            TcSession -> noBackend
                             _         -> backend dflags
+            , ghcLink = case session_type of
+                            TcSession -> NoLink
+                            _         -> ghcLink dflags
             , homeUnitInstantiations_ = insts
                                      -- if we don't have any instantiation, don't
                                      -- fill `homeUnitInstanceOfId` as it makes no
                                      -- sense (we're not instantiating anything)
-            , homeUnitInstanceOf_   = if null insts then Nothing else Just (indefUnit cid)
+            , homeUnitInstanceOf_   = if null insts then Nothing else Just cid
             , homeUnitId_ = case session_type of
                 TcSession -> newUnitId cid Nothing
                 -- No hash passed if no instances
@@ -208,7 +218,7 @@ withBkpSession cid insts deps session_type do_this = do
                 -- Make sure to write interfaces when we are type-checking
                 -- indefinite packages.
                 TcSession
-                  | backend dflags /= NoBackend
+                  | backendSupportsInterfaceWriting $ backend dflags
                   -> EnumSet.insert Opt_WriteInterface (generalFlags dflags)
                 _ -> generalFlags dflags
 
@@ -243,21 +253,21 @@ withBkpSession cid insts deps session_type do_this = do
 
 withBkpExeSession :: [(Unit, ModRenaming)] -> BkpM a -> BkpM a
 withBkpExeSession deps do_this =
-    withBkpSession (Indefinite (UnitId (fsLit "main"))) [] deps ExeSession do_this
+    withBkpSession (UnitId (fsLit "main")) [] deps ExeSession do_this
 
-getSource :: IndefUnitId -> BkpM (LHsUnit HsComponentId)
+getSource :: UnitId -> BkpM (LHsUnit HsComponentId)
 getSource cid = do
     bkp_env <- getBkpEnv
     case Map.lookup cid (bkp_table bkp_env) of
         Nothing -> pprPanic "missing needed dependency" (ppr cid)
         Just lunit -> return lunit
 
-typecheckUnit :: IndefUnitId -> [(ModuleName, Module)] -> BkpM ()
+typecheckUnit :: UnitId -> [(ModuleName, Module)] -> BkpM ()
 typecheckUnit cid insts = do
     lunit <- getSource cid
     buildUnit TcSession cid insts lunit
 
-compileUnit :: IndefUnitId -> [(ModuleName, Module)] -> BkpM ()
+compileUnit :: UnitId -> [(ModuleName, Module)] -> BkpM ()
 compileUnit cid insts = do
     -- Let everyone know we're building this unit
     msgUnitId (mkVirtUnit cid insts)
@@ -285,7 +295,7 @@ hsunitDeps include_sigs unit = concatMap get_dep (hsunitBody unit)
             convRn (L _ (Renaming (L _ from) (Just (L _ to)))) = (from, to)
     get_dep _ = []
 
-buildUnit :: SessionType -> IndefUnitId -> [(ModuleName, Module)] -> LHsUnit HsComponentId -> BkpM ()
+buildUnit :: SessionType -> UnitId -> [(ModuleName, Module)] -> LHsUnit HsComponentId -> BkpM ()
 buildUnit session cid insts lunit = do
     -- NB: include signature dependencies ONLY when typechecking.
     -- If we're compiling, it's not necessary to recursively
@@ -318,10 +328,10 @@ buildUnit session cid insts lunit = do
     conf <- withBkpSession cid insts deps_w_rns session $ do
 
         dflags <- getDynFlags
-        mod_graph <- hsunitModuleGraph (unLoc lunit)
+        mod_graph <- hsunitModuleGraph False (unLoc lunit)
 
         msg <- mkBackpackMsg
-        ok <- load' LoadAllTargets (Just msg) mod_graph
+        ok <- load' noIfaceCache LoadAllTargets mkUnknownDiagnostic (Just msg) mod_graph
         when (failed ok) (liftIO $ exitWith (ExitFailure 1))
 
         let hi_dir = expectJust (panic "hiDir Backpack") $ hiDir dflags
@@ -333,14 +343,13 @@ buildUnit session cid insts lunit = do
         -- Compile relevant only
         hsc_env <- getSession
         let home_mod_infos = eltsUDFM (hsc_HPT hsc_env)
-            linkables = map (expectJust "bkp link" . hm_linkable)
+            linkables = map (expectJust "bkp link" . homeModInfoObject)
                       . filter ((==HsSrcFile) . mi_hsc_src . hm_iface)
                       $ home_mod_infos
-            getOfiles (LM _ _ us) = map nameOfObject (filter isObject us)
-            obj_files = concatMap getOfiles linkables
+            obj_files = concatMap linkableFiles linkables
             state     = hsc_units hsc_env
 
-        let compat_fs = unitIdFS (indefUnit cid)
+        let compat_fs = unitIdFS cid
             compat_pn = PackageName compat_fs
             unit_id   = homeUnitId (hsc_home_unit hsc_env)
 
@@ -408,9 +417,9 @@ compileExe lunit = do
     forM_ (zip [1..] deps) $ \(i, dep) ->
         compileInclude (length deps) (i, dep)
     withBkpExeSession deps_w_rns $ do
-        mod_graph <- hsunitModuleGraph (unLoc lunit)
+        mod_graph <- hsunitModuleGraph True (unLoc lunit)
         msg <- mkBackpackMsg
-        ok <- load' LoadAllTargets (Just msg) mod_graph
+        ok <- load' noIfaceCache LoadAllTargets mkUnknownDiagnostic (Just msg) mod_graph
         when (failed ok) (liftIO $ exitWith (ExitFailure 1))
 
 -- | Register a new virtual unit database containing a single unit
@@ -419,7 +428,8 @@ addUnit u = do
     hsc_env <- getSession
     logger <- getLogger
     let dflags0 = hsc_dflags hsc_env
-    newdbs <- case hsc_unit_dbs hsc_env of
+    let old_unit_env = hsc_unit_env hsc_env
+    newdbs <- case ue_unit_dbs old_unit_env of
         Nothing  -> panic "addUnit: called too early"
         Just dbs ->
          let newdb = UnitDatabase
@@ -427,22 +437,23 @@ addUnit u = do
                , unitDatabaseUnits = [u]
                }
          in return (dbs ++ [newdb]) -- added at the end because ordering matters
-    (dbs,unit_state,home_unit,mconstants) <- liftIO $ initUnits logger dflags0 (Just newdbs)
+    (dbs,unit_state,home_unit,mconstants) <- liftIO $ initUnits logger dflags0 (Just newdbs) (hsc_all_home_unit_ids hsc_env)
 
     -- update platform constants
     dflags <- liftIO $ updatePlatformConstants dflags0 mconstants
 
-    let unit_env = UnitEnv
+    let unit_env = ue_setUnits unit_state $ ue_setUnitDbs (Just dbs) $ UnitEnv
           { ue_platform  = targetPlatform dflags
           , ue_namever   = ghcNameVersion dflags
-          , ue_home_unit = home_unit
-          , ue_units     = unit_state
+          , ue_current_unit = homeUnitId home_unit
+
+          , ue_home_unit_graph =
+                unitEnv_singleton
+                    (homeUnitId home_unit)
+                    (mkHomeUnitEnv dflags (ue_hpt old_unit_env) (Just home_unit))
+          , ue_eps       = ue_eps old_unit_env
           }
-    setSession $ hsc_env
-        { hsc_unit_dbs = Just dbs
-        , hsc_dflags   = dflags
-        , hsc_unit_env = unit_env
-        }
+    setSession $ hscSetFlags dflags $ hsc_env { hsc_unit_env = unit_env }
 
 compileInclude :: Int -> (Int, Unit) -> BkpM ()
 compileInclude n (i, uid) = do
@@ -473,7 +484,7 @@ data BkpEnv
         -- | The filename of the bkp file we're compiling
         bkp_filename :: FilePath,
         -- | Table of source units which we know how to compile
-        bkp_table :: Map IndefUnitId (LHsUnit HsComponentId),
+        bkp_table :: Map UnitId (LHsUnit HsComponentId),
         -- | When a package we are compiling includes another package
         -- which has not been compiled, we bump the level and compile
         -- that.
@@ -515,13 +526,13 @@ innerBkpM do_this =
 updateEpsGhc_ :: GhcMonad m => (ExternalPackageState -> ExternalPackageState) -> m ()
 updateEpsGhc_ f = do
     hsc_env <- getSession
-    liftIO $ atomicModifyIORef' (hsc_EPS hsc_env) (\x -> (f x, ()))
+    liftIO $ atomicModifyIORef' (euc_eps (ue_eps (hsc_unit_env hsc_env))) (\x -> (f x, ()))
 
 -- | Get the EPS from a 'GhcMonad'.
 getEpsGhc :: GhcMonad m => m ExternalPackageState
 getEpsGhc = do
     hsc_env <- getSession
-    liftIO $ readIORef (hsc_EPS hsc_env)
+    liftIO $ hscEPS hsc_env
 
 -- | Run 'BkpM' in 'Ghc'.
 initBkpM :: FilePath -> [LHsUnit HsComponentId] -> BkpM a -> Ghc a
@@ -540,10 +551,10 @@ initBkpM file bkp m =
 
 -- | Print a compilation progress message, but with indentation according
 -- to @level@ (for nested compilation).
-backpackProgressMsg :: Int -> Logger -> DynFlags -> SDoc -> IO ()
-backpackProgressMsg level logger dflags msg =
-    compilationProgressMsg logger dflags $ text (replicate (level * 2) ' ') -- TODO: use GHC.Utils.Ppr.RStr
-                                      <> msg
+backpackProgressMsg :: Int -> Logger -> SDoc -> IO ()
+backpackProgressMsg level logger msg =
+    compilationProgressMsg logger $ text (replicate (level * 2) ' ') -- TODO: use GHC.Utils.Ppr.RStr
+                                    <> msg
 
 -- | Creates a 'Messager' for Backpack compilation; this is basically
 -- a carbon copy of 'batchMsg' but calling 'backpackProgressMsg', which
@@ -556,25 +567,28 @@ mkBackpackMsg = do
           logger = hsc_logger hsc_env
           state = hsc_units hsc_env
           showMsg msg reason =
-            backpackProgressMsg level logger dflags $ pprWithUnitState state $
+            backpackProgressMsg level logger $ pprWithUnitState state $
                 showModuleIndex mod_index <>
                 msg <> showModMsg dflags (recompileRequired recomp) node
                     <> reason
       in case node of
-        InstantiationNode _ ->
+        InstantiationNode _ _ ->
           case recomp of
-            MustCompile -> showMsg (text "Instantiating ") empty
             UpToDate
               | verbosity (hsc_dflags hsc_env) >= 2 -> showMsg (text "Skipping  ") empty
               | otherwise -> return ()
-            RecompBecause reason -> showMsg (text "Instantiating ") (text " [" <> text reason <> text "]")
-        ModuleNode _ ->
+            NeedsRecompile reason0 -> showMsg (text "Instantiating ") $ case reason0 of
+              MustCompile -> empty
+              RecompBecause reason -> text " [" <> pprWithUnitState state (ppr reason) <> text "]"
+        ModuleNode _ _ ->
           case recomp of
-            MustCompile -> showMsg (text "Compiling ") empty
             UpToDate
               | verbosity (hsc_dflags hsc_env) >= 2 -> showMsg (text "Skipping  ") empty
               | otherwise -> return ()
-            RecompBecause reason -> showMsg (text "Compiling ") (text " [" <> text reason <> text "]")
+            NeedsRecompile reason0 -> showMsg (text "Compiling ") $ case reason0 of
+              MustCompile -> empty
+              RecompBecause reason -> text " [" <> pprWithUnitState state (ppr reason) <> text "]"
+        LinkNode _ _ -> showMsg (text "Linking ")  empty
 
 -- | 'PprStyle' for Backpack messages; here we usually want the module to
 -- be qualified (so we can tell how it was instantiated.) But we try not
@@ -584,26 +598,26 @@ backpackStyle =
     mkUserStyle
         (QueryQualify neverQualifyNames
                       alwaysQualifyModules
-                      neverQualifyPackages) AllTheWay
+                      neverQualifyPackages
+                      alwaysPrintPromTick)
+        AllTheWay
 
 -- | Message when we initially process a Backpack unit.
 msgTopPackage :: (Int,Int) -> HsComponentId -> BkpM ()
 msgTopPackage (i,n) (HsComponentId (PackageName fs_pn) _) = do
-    dflags <- getDynFlags
     logger <- getLogger
     level <- getBkpLevel
-    liftIO . backpackProgressMsg level logger dflags
+    liftIO . backpackProgressMsg level logger
         $ showModuleIndex (i, n) <> text "Processing " <> ftext fs_pn
 
 -- | Message when we instantiate a Backpack unit.
 msgUnitId :: Unit -> BkpM ()
 msgUnitId pk = do
-    dflags <- getDynFlags
     logger <- getLogger
     hsc_env <- getSession
     level <- getBkpLevel
     let state = hsc_units hsc_env
-    liftIO . backpackProgressMsg level logger dflags
+    liftIO . backpackProgressMsg level logger
         $ pprWithUnitState state
         $ text "Instantiating "
            <> withPprStyle backpackStyle (ppr pk)
@@ -611,12 +625,11 @@ msgUnitId pk = do
 -- | Message when we include a Backpack unit.
 msgInclude :: (Int,Int) -> Unit -> BkpM ()
 msgInclude (i,n) uid = do
-    dflags <- getDynFlags
     logger <- getLogger
     hsc_env <- getSession
     level <- getBkpLevel
     let state = hsc_units hsc_env
-    liftIO . backpackProgressMsg level logger dflags
+    liftIO . backpackProgressMsg level logger
         $ pprWithUnitState state
         $ showModuleIndex (i, n) <> text "Including "
             <> withPprStyle backpackStyle (ppr uid)
@@ -630,7 +643,7 @@ type PackageNameMap a = UniqFM PackageName a
 -- to use this for anything
 unitDefines :: LHsUnit PackageName -> (PackageName, HsComponentId)
 unitDefines (L _ HsUnit{ hsunitName = L _ pn@(PackageName fs) })
-    = (pn, HsComponentId pn (Indefinite (UnitId fs)))
+    = (pn, HsComponentId pn (UnitId fs))
 
 bkpPackageNameMap :: [LHsUnit PackageName] -> PackageNameMap HsComponentId
 bkpPackageNameMap units = listToUFM (map unitDefines units)
@@ -706,79 +719,96 @@ convertHsModuleId (HsModuleId (L _ hsuid) (L _ modname)) = mkModule (convertHsCo
 --
 -- We don't bother trying to support GHC.Driver.Make for now, it's more trouble
 -- than it's worth for inline modules.
-hsunitModuleGraph :: HsUnit HsComponentId -> BkpM ModuleGraph
-hsunitModuleGraph unit = do
+hsunitModuleGraph :: Bool -> HsUnit HsComponentId -> BkpM ModuleGraph
+hsunitModuleGraph do_link unit = do
     hsc_env <- getSession
 
     let decls = hsunitBody unit
         pn = hsPackageName (unLoc (hsunitName unit))
         home_unit = hsc_home_unit hsc_env
 
+        sig_keys = flip map (homeUnitInstantiations home_unit) $ \(mod_name, _) -> NodeKey_Module (ModNodeKeyWithUid (GWIB mod_name NotBoot) (homeUnitId home_unit))
+        keys = [NodeKey_Module (ModNodeKeyWithUid gwib (homeUnitId home_unit)) | (DeclD hsc_src lmodname _) <- map unLoc decls, let gwib = GWIB (unLoc lmodname) (hscSourceToIsBoot hsc_src) ]
+
     --  1. Create a HsSrcFile/HsigFile summary for every
     --  explicitly mentioned module/signature.
-    let get_decl (L _ (DeclD hsc_src lmodname mb_hsmod)) =
-          Just `fmap` summariseDecl pn hsc_src lmodname mb_hsmod
+    let get_decl (L _ (DeclD hsc_src lmodname hsmod)) =
+          Just <$> summariseDecl pn hsc_src lmodname hsmod (keys ++ sig_keys)
         get_decl _ = return Nothing
-    nodes <- catMaybes `fmap` mapM get_decl decls
+    nodes <- mapMaybeM get_decl decls
 
     --  2. For each hole which does not already have an hsig file,
     --  create an "empty" hsig file to induce compilation for the
     --  requirement.
     let hsig_set = Set.fromList
           [ ms_mod_name ms
-          | ExtendedModSummary { emsModSummary = ms } <- nodes
+          | ModuleNode _ ms <- nodes
           , ms_hsc_src ms == HsigFile
           ]
     req_nodes <- fmap catMaybes . forM (homeUnitInstantiations home_unit) $ \(mod_name, _) ->
         if Set.member mod_name hsig_set
             then return Nothing
-            else fmap (Just . extendModSummaryNoDeps) $ summariseRequirement pn mod_name
-            -- Using extendModSummaryNoDeps here is okay because we're making a leaf node
-            -- representing a signature that can't depend on any other unit.
+            else fmap Just $ summariseRequirement pn mod_name
+
+    let graph_nodes = nodes ++ req_nodes ++ (instantiationNodes (homeUnitId $ hsc_home_unit hsc_env) (hsc_units hsc_env))
+        key_nodes = map mkNodeKey graph_nodes
+        all_nodes = graph_nodes ++ [LinkNode key_nodes (homeUnitId $ hsc_home_unit hsc_env) | do_link]
+    -- This error message is not very good but .bkp mode is just for testing so
+    -- better to be direct rather than pretty.
+    when
+      (length key_nodes /= length (ordNub key_nodes))
+      (pprPanic "Duplicate nodes keys in backpack file" (ppr key_nodes))
 
     -- 3. Return the kaboodle
-    return $ mkModuleGraph' $
-      (ModuleNode <$> (nodes ++ req_nodes)) ++ instantiationNodes (hsc_units hsc_env)
+    return $ mkModuleGraph $ all_nodes
 
-summariseRequirement :: PackageName -> ModuleName -> BkpM ModSummary
+
+summariseRequirement :: PackageName -> ModuleName -> BkpM ModuleGraphNode
 summariseRequirement pn mod_name = do
     hsc_env <- getSession
     let dflags = hsc_dflags hsc_env
+    let home_unit = hsc_home_unit hsc_env
+    let fopts = initFinderOpts dflags
 
     let PackageName pn_fs = pn
-    location <- liftIO $ mkHomeModLocation2 dflags mod_name
-                 (unpackFS pn_fs </> moduleNameSlashes mod_name) "hsig"
+    let location = mkHomeModLocation2 fopts mod_name
+                    (unsafeEncodeUtf $ unpackFS pn_fs </> moduleNameSlashes mod_name) (os "hsig")
 
     env <- getBkpEnv
-    time <- liftIO $ getModificationUTCTime (bkp_filename env)
+    src_hash <- liftIO $ getFileHash (bkp_filename env)
     hi_timestamp <- liftIO $ modificationTimeIfExists (ml_hi_file location)
     hie_timestamp <- liftIO $ modificationTimeIfExists (ml_hie_file location)
     let loc = srcLocSpan (mkSrcLoc (mkFastString (bkp_filename env)) 1 1)
 
-    mod <- liftIO $ addHomeModuleToFinder hsc_env mod_name location
+    let fc = hsc_FC hsc_env
+    mod <- liftIO $ addHomeModuleToFinder fc home_unit (notBoot mod_name) location
 
     extra_sig_imports <- liftIO $ findExtraSigImports hsc_env HsigFile mod_name
 
-    return ModSummary {
+    let ms = ModSummary {
         ms_mod = mod,
         ms_hsc_src = HsigFile,
         ms_location = location,
-        ms_hs_date = time,
+        ms_hs_hash = src_hash,
         ms_obj_date = Nothing,
+        ms_dyn_obj_date = Nothing,
         ms_iface_date = hi_timestamp,
         ms_hie_date = hie_timestamp,
         ms_srcimps = [],
-        ms_textual_imps = extra_sig_imports,
+        ms_textual_imps = ((,) NoPkgQual . noLoc) <$> extra_sig_imports,
+        ms_ghc_prim_import = False,
         ms_parsed_mod = Just (HsParsedModule {
                 hpm_module = L loc (HsModule {
-                        hsmodAnn = noAnn,
-                        hsmodLayout = NoLayoutInfo,
+                        hsmodExt = XModulePs {
+                            hsmodAnn = noAnn,
+                            hsmodLayout = EpNoLayout,
+                            hsmodDeprecMessage = Nothing,
+                            hsmodHaddockModHeader = Nothing
+                                             },
                         hsmodName = Just (L (noAnnSrcSpan loc) mod_name),
                         hsmodExports = Nothing,
                         hsmodImports = [],
-                        hsmodDecls = [],
-                        hsmodDeprecMessage = Nothing,
-                        hsmodHaddockModHeader = Nothing
+                        hsmodDecls = []
                     }),
                 hpm_src_files = []
             }),
@@ -786,39 +816,29 @@ summariseRequirement pn mod_name = do
         ms_hspp_opts = dflags,
         ms_hspp_buf = Nothing
         }
+    let nodes = [NodeKey_Module (ModNodeKeyWithUid (GWIB mn NotBoot) (homeUnitId home_unit)) | mn <- extra_sig_imports ]
+    return (ModuleNode nodes ms)
 
 summariseDecl :: PackageName
               -> HscSource
               -> Located ModuleName
-              -> Maybe (Located HsModule)
-              -> BkpM ExtendedModSummary
-summariseDecl pn hsc_src (L _ modname) (Just hsmod) = hsModuleToModSummary pn hsc_src modname hsmod
-summariseDecl _pn hsc_src lmodname@(L loc modname) Nothing
-    = do hsc_env <- getSession
-         -- TODO: this looks for modules in the wrong place
-         r <- liftIO $ summariseModule hsc_env
-                         emptyModNodeMap -- GHC API recomp not supported
-                         (hscSourceToIsBoot hsc_src)
-                         lmodname
-                         True -- Target lets you disallow, but not here
-                         Nothing -- GHC API buffer support not supported
-                         [] -- No exclusions
-         case r of
-            Nothing -> throwOneError (mkPlainMsgEnvelope loc (text "module" <+> ppr modname <+> text "was not found"))
-            Just (Left err) -> throwErrors err
-            Just (Right summary) -> return summary
+              -> Located (HsModule GhcPs)
+              -> [NodeKey]
+              -> BkpM ModuleGraphNode
+summariseDecl pn hsc_src (L _ modname) hsmod home_keys = hsModuleToModSummary home_keys pn hsc_src modname hsmod
 
 -- | Up until now, GHC has assumed a single compilation target per source file.
 -- Backpack files with inline modules break this model, since a single file
 -- may generate multiple output files.  How do we decide to name these files?
 -- Should there only be one output file? This function our current heuristic,
 -- which is we make a "fake" module and use that.
-hsModuleToModSummary :: PackageName
+hsModuleToModSummary :: [NodeKey]
+                     -> PackageName
                      -> HscSource
                      -> ModuleName
-                     -> Located HsModule
-                     -> BkpM ExtendedModSummary
-hsModuleToModSummary pn hsc_src modname
+                     -> Located (HsModule GhcPs)
+                     -> BkpM ModuleGraphNode
+hsModuleToModSummary home_keys pn hsc_src modname
                      hsmod = do
     let imps = hsmodImports (unLoc hsmod)
         loc  = getLoc hsmod
@@ -827,26 +847,25 @@ hsModuleToModSummary pn hsc_src modname
     -- Use the PACKAGE NAME to find the location
     let PackageName unit_fs = pn
         dflags = hsc_dflags hsc_env
+        fopts = initFinderOpts dflags
     -- Unfortunately, we have to define a "fake" location in
     -- order to appease the various code which uses the file
     -- name to figure out where to put, e.g. object files.
     -- To add insult to injury, we don't even actually use
     -- these filenames to figure out where the hi files go.
     -- A travesty!
-    location0 <- liftIO $ mkHomeModLocation2 dflags modname
-                             (unpackFS unit_fs </>
+    let location0 = mkHomeModLocation2 fopts modname
+                             (unsafeEncodeUtf $ unpackFS unit_fs </>
                               moduleNameSlashes modname)
                               (case hsc_src of
-                                HsigFile -> "hsig"
-                                HsBootFile -> "hs-boot"
-                                HsSrcFile -> "hs")
+                                HsigFile   -> os "hsig"
+                                HsBootFile -> os "hs-boot"
+                                HsSrcFile  -> os "hs")
     -- DANGEROUS: bootifying can POISON the module finder cache
     let location = case hsc_src of
                         HsBootFile -> addBootSuffixLocnOut location0
                         _ -> location0
     -- This duplicates a pile of logic in GHC.Driver.Make
-    env <- getBkpEnv
-    time <- liftIO $ getModificationUTCTime (bkp_filename env)
     hi_timestamp <- liftIO $ modificationTimeIfExists (ml_hi_file location)
     hie_timestamp <- liftIO $ modificationTimeIfExists (ml_hie_file location)
 
@@ -854,13 +873,16 @@ hsModuleToModSummary pn hsc_src modname
     let (src_idecls, ord_idecls) = partition ((== IsBoot) . ideclSource . unLoc) imps
 
              -- GHC.Prim doesn't exist physically, so don't go looking for it.
-        ordinary_imps = filter ((/= moduleName gHC_PRIM) . unLoc . ideclName . unLoc)
-                               ord_idecls
+        (ordinary_imps, ghc_prim_import)
+          = partition ((/= moduleName gHC_PRIM) . unLoc . ideclName . unLoc)
+              ord_idecls
 
         implicit_prelude = xopt LangExt.ImplicitPrelude dflags
         implicit_imports = mkPrelImports modname loc
                                          implicit_prelude imps
-        convImport (L _ i) = (fmap sl_fs (ideclPkgQual i), reLoc $ ideclName i)
+
+        rn_pkg_qual = renameRawPkgQual (hsc_unit_env hsc_env) modname
+        convImport (L _ i) = (rn_pkg_qual (ideclPkgQual i), reLoc $ ideclName i)
 
     extra_sig_imports <- liftIO $ findExtraSigImports hsc_env hsc_src modname
 
@@ -868,10 +890,11 @@ hsModuleToModSummary pn hsc_src modname
     (implicit_sigs, inst_deps) <- liftIO $ implicitRequirementsShallow hsc_env normal_imports
 
     -- So that Finder can find it, even though it doesn't exist...
-    this_mod <- liftIO $ addHomeModuleToFinder hsc_env modname location
-    return $ ExtendedModSummary
-      { emsModSummary =
-          ModSummary {
+    this_mod <- liftIO $ do
+      let home_unit = hsc_home_unit hsc_env
+      let fc        = hsc_FC hsc_env
+      addHomeModuleToFinder fc home_unit (GWIB modname (hscSourceToIsBoot hsc_src)) location
+    let ms = ModSummary {
             ms_mod = this_mod,
             ms_hsc_src = hsc_src,
             ms_location = location,
@@ -881,28 +904,42 @@ hsModuleToModSummary pn hsc_src modname
             ms_hspp_opts = dflags,
             ms_hspp_buf = Nothing,
             ms_srcimps = map convImport src_idecls,
+            ms_ghc_prim_import = not (null ghc_prim_import),
             ms_textual_imps = normal_imports
                            -- We have to do something special here:
                            -- due to merging, requirements may end up with
                            -- extra imports
-                           ++ extra_sig_imports
-                           ++ ((,) Nothing . noLoc <$> implicit_sigs),
+                           ++ ((,) NoPkgQual . noLoc <$> extra_sig_imports)
+                           ++ ((,) NoPkgQual . noLoc <$> implicit_sigs),
             -- This is our hack to get the parse tree to the right spot
             ms_parsed_mod = Just (HsParsedModule {
                     hpm_module = hsmod,
                     hpm_src_files = [] -- TODO if we preprocessed it
                 }),
-            ms_hs_date = time,
+            -- Source hash = fingerprint0, so the recompilation tests do not recompile
+            -- too much. In future, if necessary then could get the hash by just hashing the
+            -- relevant part of the .bkp file.
+            ms_hs_hash = fingerprint0,
             ms_obj_date = Nothing, -- TODO do this, but problem: hi_timestamp is BOGUS
+            ms_dyn_obj_date = Nothing, -- TODO do this, but problem: hi_timestamp is BOGUS
             ms_iface_date = hi_timestamp,
             ms_hie_date = hie_timestamp
           }
-      , emsInstantiatedUnits = inst_deps
-      }
+
+    -- Now, what are the dependencies.
+    let inst_nodes = map NodeKey_Unit inst_deps
+        mod_nodes  =
+          -- hs-boot edge
+          [k | k <- [NodeKey_Module (ModNodeKeyWithUid (GWIB (ms_mod_name ms) IsBoot)  (moduleUnitId this_mod))], NotBoot == isBootSummary ms,  k `elem` home_keys ] ++
+          -- Normal edges
+          [k | (_, mnwib) <- msDeps ms, let k = NodeKey_Module (ModNodeKeyWithUid (fmap unLoc mnwib) (moduleUnitId this_mod)), k `elem` home_keys]
+
+
+    return (ModuleNode (mod_nodes ++ inst_nodes) ms)
 
 -- | Create a new, externally provided hashed unit id from
 -- a hash.
-newUnitId :: IndefUnitId -> Maybe FastString -> UnitId
+newUnitId :: UnitId -> Maybe FastString -> UnitId
 newUnitId uid mhash = case mhash of
-   Nothing   -> indefUnit uid
-   Just hash -> UnitId (unitIdFS (indefUnit uid) `appendFS` mkFastString "+" `appendFS` hash)
+   Nothing   -> uid
+   Just hash -> UnitId (concatFS [unitIdFS uid, fsLit "+", hash])
