@@ -126,9 +126,7 @@ packageArgs = do
         -------------------------------- ghcPrim -------------------------------
         , package ghcPrim ? mconcat
           [ builder (Cabal Flags) ? flag NeedLibatomic `cabalFlag` "need-atomic"
-
-          , builder (Cc CompileC) ? (not <$> flag CcLlvmBackend) ?
-            input "**/cbits/atomic.c"  ? arg "-Wno-sync-nand" ]
+          ]
 
         -------------------------------- ghcBoot ------------------------------
         , package ghcBoot ?
@@ -435,6 +433,9 @@ rtsPackageArgs = package rts ? do
 
           , input "**/RetainerProfile.c" ? flag CcLlvmBackend ?
             arg "-Wno-incompatible-pointer-types"
+
+          , input "**/prim/atomic.c"  ? (not <$> flag CcLlvmBackend) ?
+            arg "-Wno-sync-nand"
           ]
 
     mconcat
