@@ -141,7 +141,7 @@ module Language.Haskell.TH.Lib (
 
    ) where
 
-import GHC.Internal.TH.Lib hiding
+import GHC.Boot.TH.Lib hiding
   ( tySynD
   , dataD
   , newtypeD
@@ -181,7 +181,7 @@ import GHC.Internal.TH.Lib hiding
   , Role
   , InjectivityAnn
   )
-import qualified GHC.Internal.TH.Lib as Internal
+import qualified GHC.Boot.TH.Lib as Internal
 import Language.Haskell.TH.Syntax
 
 import Control.Applicative (Applicative(..))
