@@ -472,6 +472,7 @@ extern char **environ;
       RTS_PROF_SYMBOLS                                                  \
       RTS_LIBDW_SYMBOLS                                                 \
       SymI_HasProto(StgReturn)                                          \
+      SymI_HasDataProto(ghc_hs_iface)                                       \
       SymI_HasDataProto(stg_gc_noregs)                                      \
       SymI_HasDataProto(stg_ret_v_info)                                     \
       SymI_HasDataProto(stg_ret_p_info)                                     \
@@ -597,6 +598,7 @@ extern char **environ;
       SymI_HasProto(hs_hpc_module)                                      \
       SymI_HasProto(hs_thread_done)                                     \
       SymI_HasProto(hs_try_putmvar)                                     \
+      SymI_HasProto(hs_try_putmvar_with_value)                          \
       SymI_HasProto(defaultRtsConfig)                                   \
       SymI_HasProto(initLinker)                                         \
       SymI_HasProto(initLinker_)                                        \
@@ -1015,6 +1017,116 @@ extern char **environ;
 #define RTS_FINI_ARRAY_SYMBOLS
 #endif
 
+#define SymI_HasProtoAllSizes(symbol)                \
+      SymI_HasProto(symbol##8)                       \
+      SymI_HasProto(symbol##16)                      \
+      SymI_HasProto(symbol##32)                      \
+      SymI_HasProto(symbol##64)
+
+#if !defined(arm_HOST_ARCH)
+#define RTS_ATOMICS_SYMBOLS                          \
+      SymI_HasProtoAllSizes(hs_atomic_add)           \
+      SymI_HasProtoAllSizes(hs_atomic_sub)           \
+      SymI_HasProtoAllSizes(hs_atomic_and)           \
+      SymI_HasProtoAllSizes(hs_atomic_nand)          \
+      SymI_HasProtoAllSizes(hs_atomic_or)            \
+      SymI_HasProtoAllSizes(hs_atomic_xor)           \
+      SymI_HasProtoAllSizes(hs_cmpxchg)              \
+      SymI_HasProtoAllSizes(hs_xchg)                 \
+      SymI_HasProtoAllSizes(hs_atomicread)           \
+      SymI_HasProtoAllSizes(hs_atomicwrite)
+#else
+// No atomics on arm32. See e9abcad4cc3
+#define RTS_ATOMICS_SYMBOLS
+#endif
+
+// In rts/longlong.c
+#if WORD_SIZE_IN_BITS < 64
+#define RTS_SYMBOLS_LONGLONG                        \
+      SymI_HasProto(hs_eq64)                        \
+      SymI_HasProto(hs_ne64)                        \
+      SymI_HasProto(hs_gtWord64)                    \
+      SymI_HasProto(hs_geWord64)                    \
+      SymI_HasProto(hs_ltWord64)                    \
+      SymI_HasProto(hs_leWord64)                    \
+      SymI_HasProto(hs_gtInt64)                     \
+      SymI_HasProto(hs_geInt64)                     \
+      SymI_HasProto(hs_ltInt64)                     \
+      SymI_HasProto(hs_leInt64)                     \
+      SymI_HasProto(hs_neg64)                       \
+      SymI_HasProto(hs_add64)                       \
+      SymI_HasProto(hs_sub64)                       \
+      SymI_HasProto(hs_mul64)                       \
+      SymI_HasProto(hs_remWord64)                   \
+      SymI_HasProto(hs_quotWord64)                  \
+      SymI_HasProto(hs_remInt64)                    \
+      SymI_HasProto(hs_quotInt64)                   \
+      SymI_HasProto(hs_and64)                       \
+      SymI_HasProto(hs_or64)                        \
+      SymI_HasProto(hs_xor64)                       \
+      SymI_HasProto(hs_not64)                       \
+      SymI_HasProto(hs_uncheckedShiftL64)           \
+      SymI_HasProto(hs_uncheckedShiftRL64)          \
+      SymI_HasProto(hs_uncheckedIShiftRA64)         \
+      SymI_HasProto(hs_intToInt64)                  \
+      SymI_HasProto(hs_int64ToInt)                  \
+      SymI_HasProto(hs_wordToWord64)                \
+      SymI_HasProto(hs_word64ToWord)
+#else
+#define RTS_SYMBOLS_LONGLONG
+#endif
+
+// rts/prim/vectorQuotRem.c and rts/prim/int64x2minmax
+#if defined(__SSE2__)
+#define RTS_SYMBOLS_VECTORQUOTREM                    \
+      SymI_HasProto(hs_quotInt8X16)                  \
+      SymI_HasProto(hs_quotInt16X8)                  \
+      SymI_HasProto(hs_quotInt32X4)                  \
+      SymI_HasProto(hs_quotInt64X2)                  \
+      SymI_HasProto(hs_quotWord8X16)                 \
+      SymI_HasProto(hs_quotWord16X8)                 \
+      SymI_HasProto(hs_quotWord32X4)                 \
+      SymI_HasProto(hs_quotWord64X2)                 \
+      SymI_HasProto(hs_remInt8X16)                   \
+      SymI_HasProto(hs_remInt16X8)                   \
+      SymI_HasProto(hs_remInt32X4)                   \
+      SymI_HasProto(hs_remInt64X2)                   \
+      SymI_HasProto(hs_remWord8X16)                  \
+      SymI_HasProto(hs_remWord16X8)                  \
+      SymI_HasProto(hs_remWord32X4)                  \
+      SymI_HasProto(hs_remWord64X2)
+#define RTS_SYMBOLS_INT64X2MINMAX                    \
+      SymI_HasProto(hs_minInt64X2)                   \
+      SymI_HasProto(hs_maxInt64X2)                   \
+      SymI_HasProto(hs_minWord64X2)                  \
+      SymI_HasProto(hs_maxWord64X2)
+#else
+#define RTS_SYMBOLS_VECTORQUOTREM
+#define RTS_SYMBOLS_INT64X2MINMAX
+#endif
+
+// Symbols on files in rts/prim/*
+#define RTS_SYMBOLS_PRIM                             \
+      RTS_ATOMICS_SYMBOLS                            \
+      RTS_SYMBOLS_INT64X2MINMAX                      \
+      RTS_SYMBOLS_LONGLONG                           \
+      RTS_SYMBOLS_VECTORQUOTREM                      \
+      SymI_HasProtoAllSizes(hs_bitrev)               \
+      SymI_HasProto(hs_bswap16)                      \
+      SymI_HasProto(hs_bswap32)                      \
+      SymI_HasProto(hs_bswap64)                      \
+      SymI_HasProtoAllSizes(hs_clz)                  \
+      SymI_HasProtoAllSizes(hs_ctz)                  \
+      SymI_HasProto(hs_mulIntMayOflo)                \
+      SymI_HasProtoAllSizes(hs_pdep)                 \
+      SymI_HasProtoAllSizes(hs_pext)                 \
+      SymI_HasProtoAllSizes(hs_pext)                 \
+      SymI_HasProto(hs_popcnt)                       \
+      SymI_HasProtoAllSizes(hs_popcnt)               \
+      SymI_HasProto(hs_word2float32)                 \
+      SymI_HasProto(hs_word2float64)
+
+
 /* entirely bogus claims about types of these symbols */
 #define SymI_NeedsProto(vvv)  extern void vvv(void);
 #define SymI_NeedsDataProto(vvv)  extern StgWord vvv[];
@@ -1048,6 +1160,7 @@ RTS_ARCH_LIBGCC_SYMBOLS
 RTS_FINI_ARRAY_SYMBOLS
 RTS_LIBFFI_SYMBOLS
 RTS_ARM_OUTLINE_ATOMIC_SYMBOLS
+RTS_SYMBOLS_PRIM
 
 #undef SymI_NeedsProto
 #undef SymI_NeedsDataProto
@@ -1091,6 +1204,7 @@ RtsSymbolVal rtsSyms[] = {
       RTS_FINI_ARRAY_SYMBOLS
       RTS_LIBFFI_SYMBOLS
       RTS_ARM_OUTLINE_ATOMIC_SYMBOLS
+      RTS_SYMBOLS_PRIM
       SymI_HasDataProto(nonmoving_write_barrier_enabled)
       { 0, 0, STRENGTH_NORMAL, SYM_TYPE_CODE } /* sentinel */
 };
